//! This module contains the whole engine that manages everything.

pub mod bbox;
pub mod character;
pub mod controls;
pub mod event;
pub mod map;
pub mod math;
pub mod physics;
pub mod scene;
pub mod texture;
pub mod vector;

use std::cell::RefCell;
use std::rc::Rc;
use std::time::Duration;

use wasm_bindgen::prelude::*;
use wasm_bindgen::JsCast;

use crate::engine::bbox::Bbox;
use crate::engine::character::Character;
use crate::engine::controls::Controls;
use crate::engine::event::{Key, Keyboard};
use crate::engine::map::Map;
use crate::engine::scene::{Scene, State};
use crate::engine::texture::{Texture, TextureManager};
use crate::engine::vector::Vector;
use crate::{error_js, log, Result};

macro_rules! unwrap {
    ($t: expr) => {{
        match $t {
            None => return Err(JsValue::from_str("cannot unwrap on none")),
            Some(x) => x,
        }
    }};
}

/// Our game engine.
#[derive(Clone)]
pub struct Engine {
    /// The inner engine.
    ///
    /// We need Rc<RefCell> in order to deal with events.
    pub inner: Rc<RefCell<InnerEngine>>,

    /// The web page window.
    pub window: Rc<web_sys::Window>,

    /// The web page document.
    pub document: Rc<web_sys::Document>,

    /// The canvas rendering context.
    ///
    /// We keep a reference so that we can easily render things.
    pub context: Rc<web_sys::CanvasRenderingContext2d>,
}

impl Engine {
    /// Creates a new engine.
    pub fn new() -> Result<Engine> {
        let window = unwrap!(web_sys::window());
        let document = unwrap!(window.document());
        let canvas = unwrap!(document.get_element_by_id("canvas"));
        let canvas: web_sys::HtmlCanvasElement = canvas.dyn_into::<web_sys::HtmlCanvasElement>()?;

        canvas.set_width(1920);
        canvas.set_height(1080);

        let context =
            unwrap!(canvas.get_context("2d")?).dyn_into::<web_sys::CanvasRenderingContext2d>()?;

        let map = Map::from_str(include_str!("../../static/levels/level1.lvl")).unwrap();
        let mut scene = Scene::from_map(map);

        let character = Character::with_controls(Controls::default_keyboard());
        scene.add(character);

        let inner = InnerEngine::new(&document, scene)?;

        let window = Rc::new(window);
        let document = Rc::new(document);
        let context = Rc::new(context);

        Ok(Engine {
            inner: Rc::new(RefCell::new(inner)),
            window,
            document,
            context,
        })
    }

    /// Starts the engine.
    pub fn start(&self) -> Result<()> {
        let clone = self.clone();
        let cb = Closure::<dyn FnMut(_)>::new(move |_event: web_sys::Event| {
            if let Err(e) = clone.run() {
                error_js!(e);
            }
        });

        (*self.window).request_animation_frame(cb.as_ref().unchecked_ref())?;

        cb.forget();

        Ok(())
    }

    /// Launches a loop of the engine, and schedules the next one.
    pub fn run(&self) -> Result<()> {
        // Perform update
        self.update()?;

        // Perform render
        self.render()?;

        // Schedule next render
        let clone = self.clone();
        let cb = Closure::<dyn FnMut(_)>::new(move |_event: web_sys::Event| {
            if let Err(e) = clone.run() {
                error_js!(e);
            }
        });

        (*self.window).request_animation_frame(cb.as_ref().unchecked_ref())?;

        cb.forget();

        Ok(())
    }

    /// Triggers the update of the model of the engine.
    pub fn update(&self) -> Result<()> {
        let mut inner = self.inner.borrow_mut();

        // Manage events
        while let Some(event) = inner.keyboard.pop() {
            inner.scene.manage_event(&event);
        }

        // Manage the physics
        let duration = Duration::from_millis(60);

        let keyboard = inner.keyboard.clone();
        if inner.scene.update(&duration, &keyboard) == State::Finished {
            // running = false;
        }

        Ok(())
    }

    /// Draw a drawable.
    pub fn draw<D: Drawable>(&self, drawable: &D) -> Result<()> {
        let inner = self.inner.borrow();
        let image = inner.textures.get(drawable.texture());

        let source = drawable.texture_rect();
        let mut dest = source.clone();
        dest.position = drawable.position();

        image.render(source, dest, &self.context)?;
        Ok(())
    }

    /// Performs the rendering of the engine.
    pub fn render(&self) -> Result<()> {
        let inner = self.inner.borrow();

        // Clear render
        self.context.clear_rect(0.0, 0.0, 1920.0, 1080.0);
        self.context
            .set_fill_style(&JsValue::from_str("rgb(135, 206, 235)"));
        self.context.fill_rect(0.0, 0.0, 1920.0, 1080.0);

        // Draw the scene
        let map = inner.scene.map();
        let rows = map.rows();
        let cols = map.cols();

        for i in 0..rows {
            for j in 0..cols {
                let tile = map.at(i, j);
                if tile.graphic.is_visible() {
                    self.draw(&tile)?;
                }
            }
        }

        // Draw characters
        for c in inner.scene.characters() {
            if true {
                self.draw(c)?;
            }
        }

        Ok(())
    }
}

/// The data contained in our engine.
pub struct InnerEngine {
    /// The scene of the engine.
    pub scene: Scene,

    /// The keyboard.
    pub keyboard: Keyboard,

    /// The texture manager.
    pub textures: TextureManager,
}

impl InnerEngine {
    /// Initializes the engine.
    pub fn new(document: &web_sys::Document, scene: Scene) -> Result<InnerEngine> {
        Ok(InnerEngine {
            scene,
            keyboard: Keyboard::new(document)?,
            textures: TextureManager::new()?,
        })
    }
}

/// Our custom drawable trait.
pub trait Drawable {
    /// Returns the texture of the drawable.
    fn texture(&self) -> Texture;

    /// Returns the coordinates to use on the texture.
    fn texture_rect(&self) -> Bbox;

    /// Returns the position on which the drawable should be drawn.
    fn position(&self) -> Vector;
}
